/* Here is some C example code that shows how to catch overflow exceptions
   on glibc/i386. You will have to turn that into a subroutine library and
   call it from FORTRAN. For more information about the FPU status word
   (in *ctx.fpstate) see the Intel manuals (available at
   http://developer.intel.com). Be careful on the flags you set in the FPU
   control register, the standard math library might break if you e.g. 
   change the rounding mode. Based on code from a Usenet post by ak@muc.de.
   T.Crane 18th March 2003 */

#include <fpu_control.h>
#include <signal.h>
#include <bits/sigcontext.h>
#include <stdio.h>

#ifndef __i386__
#error you lose
#endif

fpu_control_t fpu_cw_mask;

void enable_fpu_interrupt(fpu_control_t irqs)
{
 _FPU_GETCW(fpu_cw_mask);
 fpu_cw_mask &= ~irqs;
 _FPU_SETCW(fpu_cw_mask);
}

void wrt_fpu_status (unsigned int status)
{
  if (status & _FPU_MASK_IM) printf ("Invalid ");
  if (status & _FPU_MASK_DM) printf ("Denormalised ");
  if (status & _FPU_MASK_ZM) printf ("Division-by-zero ");
  if (status & _FPU_MASK_OM) printf ("Overflow ");
  if (status & _FPU_MASK_UM) printf ("Underflow ");
  if (status & _FPU_MASK_PM) printf ("Inexact ");
  putchar(10);
}


/* Be careful what you call in the signal handler. The FPE might have
   been caused by stdio or other library functions that are usually _not_
   reentrant. The best way to recover is a longjmp(). 
   FPU Control Word bits;

  |   15-13    12  11-10  9-8     7-6     5    4    3    2    1    0
  | reserved | IC | RC  | PC | reserved | PM | UM | OM | ZM | DM | IM
 
  IM: Invalid operation mask
  DM: Denormalized operand mask
  ZM: Zero-divide mask
  OM: Overflow mask
  UM: Underflow mask
  PM: Precision (inexact result) mask 
*/

void fpe_handler(int signum, struct sigcontext ctx)
{
  unsigned int fpu_cw, unmasked_bits;
  printf("\n**** fpu exception at 0x%x\n", ctx.eip); 
  /* printf("fpu status = %x\n", ctx.fpstate->status); */
  printf("all fpu exception status bits: ");
  wrt_fpu_status(ctx.fpstate->status);
  unmasked_bits = ctx.fpstate->status & ~fpu_cw_mask;
  printf("unmasked fpu exception status bits: ");
  wrt_fpu_status(unmasked_bits);
  exit(0);
}


int setup_fpe_handler__(int *mask)
{
 struct sigaction sa = {0};

 sa.sa_handler = (void (*)(int)) fpe_handler;
 sa.sa_flags = 0;
 sigaction(SIGFPE, &sa, NULL);

 /* enable_fpu_interrupt(_FPU_MASK_OM); turn on overflow */
 /* Clear the FPU's status bits before enabling interrupts 
    Failure to do this following a previous denormalised result, can result
    in an immediate exception */
 do_fnclex(); 
 enable_fpu_interrupt((fpu_control_t)*mask); /* turn on specified mask bits */
 return 0;
}
